/*
*  Program name:       tky_jgd  ver. 1.0
*
*  Program function:   Tokyo datum <==> JGD 2000 Datum
*
*  Programmed by:      Yasuaki MURATA,
*                      Geological Survey of Japan, Tsukuba
*                      June 9, 2003
*
*  Date of revision:
*
*  Arguments:          1. Parameter file name
*                      2. -R (JGD 2000 -> Tokyo)
*
*  Input:              JKG (unit:second) from stdin
*                        project only 'l' 'm' 'A' 'W' data
*
*  Output:             converted JKG to stdout
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define	T1_Def	-147.5
#define	T2_Def	507.3
#define	T3_Def	680.5
#define	Bessel_a	6377397.155
#define	Bessel_1f	299.152813
#define	GRS80_a 	6378137.0
#define	GRS80_1f	298.257223563
#define	PI_S	0.00000484813681109536

#define	mLons	22
#define	mLone	45
#define	mLats	36
#define	mLate	68

#define	nLon	(mLone-mLons+1)*80
#define	nLat	(mLate-mLats+1)*80

int	Par, Rev ;
double	T1, T2, T3 ;
double	Bessel_f, Bessel_e2 ;
double	GRS80_f, GRS80_e2 ;
double	dLon[nLon][nLat], dLat[nLon][nLat] ;

main(argc, argv)
int	argc ;
char	*argv[] ;
{
	FILE	*fp ;
	char	hed, line[500] ;
	long	i, lon, lat, lon1, lat1, lon2, lat2, lon3, lat3 ;
	double	L, B, H, L2, B2, H2 ;
	void	tky2jgd(), jgd2tky() ;

	Par = 1 ;
	T1 = T1_Def ;
	T2 = T2_Def ;
	T3 = T3_Def ;
	Rev = 0 ;

	Bessel_f = 1.0 / Bessel_1f ;
	Bessel_e2 = Bessel_f * (2.0 - Bessel_f) ;
	GRS80_f = 1.0 / GRS80_1f ;
	GRS80_e2 = GRS80_f * (2.0 - GRS80_f) ;

	for (i = 1; i < argc; i++) {
	  if (strcmp(argv[i], "-h") == 0) {
	    fprintf(stderr, " USAGE: tky2jgd -pf dsn -3p T1 T2 T3 -r\n") ;
	    return(1) ;
	    }
	  else if (strcmp(argv[i], "-pf") == 0) {
	    Par = 2 ;
	    for (lon = 0; lon < nLon; lon++) {
	      for (lat = 0; lat < nLat; lat++) {
	        dLon[lon][lat] = 0.0 ;
	        dLat[lon][lat] = 0.0 ;
	        }
	      }
	    if ((fp = fopen(argv[++i], "r")) == NULL) {
	      fprintf(stderr, "Parameter file is not found.\n") ;
	      exit(-1) ;
	      }
	    fgets(line, 500, fp) ;
	    fgets(line, 500, fp) ;
	    while (1) {
	      if (fgets(line, 500, fp) == NULL) break ;
	      sscanf(line, "%2ld%2ld%1ld%1ld%1ld%1ld %lf %lf",
	        &lat1, &lon1, &lat2, &lon2, &lat3, &lon3, &B, &L) ;
	      lon = (lon1 - mLons) * 80 + lon2 * 10 + lon3 ;
	      lat = (lat1 - mLats) * 80 + lat2 * 10 + lat3 ;
	      if (lon < 0 || lon >= nLon || lat < 0 || lat >= nLat) {
	        fprintf(stderr, "PAR file error. %s", line) ;
	        exit(-1) ;
	        }
	      dLon[lon][lat] = L ;
	      dLat[lon][lat] = B ;
	      }
	    fclose(fp) ;
	    fprintf(stderr, "PAR file read end\n") ;
	    }
	  else if (strcmp(argv[i], "-3p") == 0) {
	    Par = 1 ;
	    T1 = atof(argv[++i]) ;
	    T2 = atof(argv[++i]) ;
	    T3 = atof(argv[++i]) ;
	    }
	  else if (strcmp(argv[i], "-r")   == 0) Rev = 1 ;
	  else if (strcmp(argv[i], "-R")   == 0) Rev = 1 ;
	  }

	H = 0.0 ;

	while ((fgets(line, 500, stdin)) != NULL) {
	  if (line[0] == 'l' || line[0] == 'm') {
	    if (sscanf(line, "%c %lf %lf", &hed, &L, &B) != 3) {
	      fprintf(stderr, "JKG l/m data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2jgd(L, B, H, &L2, &B2, &H2) ;
	    else jgd2tky(L, B, H, &L2, &B2, &H2) ;
	    printf("%c %lf %lf\n", hed, L2, B2) ;
	    }
	  else if(line[0] == 'A' || line[0] == 'W') {
	    if (sscanf(line, "%c %lf %lf", &hed, &L, &B) != 3) {
	      fprintf(stderr, "JKG A/W data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2jgd(L, B, H, &L2, &B2, &H2) ;
	    else jgd2tky(L, B, H, &L2, &B2, &H2) ;
	    printf("%c %lf %lf", hed, L2, B2) ;
	    if (sscanf(line, "%c %*lf %*lf %lf %lf", &hed, &L, &B) != 3) {
	      fprintf(stderr, "JKG A/W data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2jgd(L, B, H, &L2, &B2, &H2) ;
	    else jgd2tky(L, B, H, &L2, &B2, &H2) ;
	    printf(" %lf %lf\n", L2, B2) ;
	    }
	  else printf("%s", line) ;
	  }
	return(0) ;
	}
/* -------------------------------------------------------------------------- */
void	tky2jgd(L, B, H, L2, B2, H2)
double	L, B, H, *L2, *B2, *H2 ;
{
	double	rB, rL, N, X, Y, Z, p, r, mu, dB, dL ;
	int	pa, getprm() ;

	pa = Par ;
top:	if (pa == 1) {
	  rB = B * PI_S ;
	  rL = L * PI_S ;
	  N = Bessel_a / sqrt(1.0 - Bessel_e2 * pow(sin(rB),2.0)) ;
	  X = (N + H) * cos(rB) * cos(rL)           + T1 ;
	  Y = (N + H) * cos(rB) * sin(rL)           + T2 ;
	  Z = (N * (1.0 - Bessel_e2) + H) * sin(rB) + T3 ;
	  *L2 = atan(Y / X) / PI_S ;
	  if (X < 0.0) *L2 = *L2 + 648000.0 ;
	  p = sqrt(X*X + Y*Y) ;
	  r = sqrt(p*p + Z*Z) ;
	  mu = atan(Z / p * ((1.0 - GRS80_f) + GRS80_e2*GRS80_a / r)) ;
	  *B2 = atan((Z*(1.0 - GRS80_f) + GRS80_e2*GRS80_a*pow(sin(mu),3.0)) /
	        ((1.0-GRS80_f) * (p - GRS80_e2*GRS80_a*pow(cos(mu),3.0)))) ;
	  *H2 = p * cos(*B2) + Z * sin(*B2) -
	        GRS80_a * sqrt(1.0 - GRS80_e2*pow(sin(*B2),2.0)) ;
	  *B2 = *B2 / PI_S ;
	  }
	else {
	  if (getprm(L, B, &dL, &dB) != 0) {
	    pa = 1 ;
	    goto top ;
	    }
	  *B2 = B + dB ;
	  *L2 = L + dL ;
	  }
	}
/* -------------------------------------------------------------------------- */
void	jgd2tky(L, B, H, L2, B2, H2)
double	L, B, H, *L2, *B2, *H2 ;
{
	double	rB, rL, N, X, Y, Z, p, r, mu, tB, tL, dB, dL ;
	int	pa, getprm() ;

	pa = Par ;
top:	if (pa == 1) {
	  rB = B * PI_S ;
	  rL = L * PI_S ;
	  N = GRS80_a / sqrt(1.0 - GRS80_e2 * pow(sin(rB),2.0)) ;
	  X = (N + H) * cos(rB) * cos(rL)          - T1 ;
	  Y = (N + H) * cos(rB) * sin(rL)          - T2 ;
	  Z = (N * (1.0 - GRS80_e2) + H) * sin(rB) - T3 ;
	  *L2 = atan(Y / X) / PI_S ;
	  if (X < 0.0) *L2 = *L2 + 648000.0 ;
	  p = sqrt(X*X + Y*Y) ;
	  r = sqrt(p*p + Z*Z) ;
	  mu = atan(Z / p * ((1.0 - Bessel_f) + Bessel_e2*Bessel_a / r)) ;
	  *B2 = atan((Z*(1.0-Bessel_f) + Bessel_e2*Bessel_a*pow(sin(mu),3.0)) /
	        ((1.0-Bessel_f)*(p - Bessel_e2*Bessel_a*pow(cos(mu),3.0)))) ;
	  *H2 = p * cos(*B2) + Z * sin(*B2) -
	        Bessel_a * sqrt(1.0 - Bessel_e2*pow(sin(*B2),2.0)) ;
	  *B2 = *B2 / PI_S ;
	  }
	else {
	  if (getprm(L,  B,  &dL, &dB) != 0) {
	    pa = 1 ;
	    goto top ;
	    }
	  tL = L - dL ;
	  tB = B - dB ;
	  if (getprm(tL, tB, &dL, &dB) != 0) {
	    pa = 1 ;
	    goto top ;
	    }
	  *L2 = L - dL ;
	  *B2 = B - dB ;
	  }
	}
/* -------------------------------------------------------------------------- */
int	getprm(L, B, dL, dB)
double	L, B, *dL, *dB ;
{
	int	lon, lat ;
	double	dlon, dlat, rlon, rlat ;
	double	v1, v2, v3, v4, va, vb ;

	dlon = L / 45.0 - (double)mLons * 80 - 8000.0 ;
	dlat = B / 30.0 - (double)mLats * 80 ;

	lon = (int)dlon ;
	lat = (int)dlat ;
	if (lon < 0 || lat < 0 || lon >= nLon || lat >= nLat) {
	  fprintf(stderr, "Area over %lf %lf %d %d\n", L, B, lon, lat) ;
	  return(-1) ;
	  }

	rlon = dlon - (double)lon ;
	rlat = dlat - (double)lat ;

	v1 = dLon[lon][lat] ;
	v2 = dLon[lon][lat+1] ;
	v3 = dLon[lon+1][lat+1] ;
	v4 = dLon[lon+1][lat] ;
	if (v1==0.0||v2==0.0||v3==0.0||v4==0.0) {
	  fprintf(stderr, "Area undef %lf %lf %d %d\n", L, B, lon, lat) ;
	  return(-1) ;
	  }
	va = (v2 - v1) * rlat + v1 ;
	vb = (v3 - v4) * rlat + v4 ;
	*dL = (vb - va) * rlon + va ;

	v1 = dLat[lon][lat] ;
	v2 = dLat[lon][lat+1] ;
	v3 = dLat[lon+1][lat+1] ;
	v4 = dLat[lon+1][lat] ;
	if (v1==0.0||v2==0.0||v3==0.0||v4==0.0) {
	  fprintf(stderr, "Area undef %lf %lf %d %d\n", L, B, lon, lat) ;
	  return(-1) ;
	  }
	va = (v2 - v1) * rlat + v1 ;
	vb = (v3 - v4) * rlat + v4 ;
	*dB = (vb - va) * rlon + va ;

	return(0) ;

	}
