/*
*  Program name:       tky_wgs  ver. 2.0
*
*  Program function:   Tokyo datum <==> WGS-84 Datum
*
*  parameter set = http://vldb.gsi-mc.go.jp/sokuchi/coordinates/localtrans.html
*
*  Programmed by:      Yasuaki MURATA,
*                      Geological Survey of Japan, Tsukuba
*                      June 9, 2003
*
*  Date of revision:
*
*  Arguments:          -pf Local-Parameter File Name   (Read the file)
*                      -3p T1 T2 T3                    (Key-in 3 Parmameters)
*                      -r (WGS 84 -> Tokyo)
*
*  Input:              JKG (unit:second) from stdin
*                        project only 'l' 'm' 'A' 'W' data
*
*  Output:             converted JKG to stdout
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define	T1_Def	-147.5
#define	T2_Def	507.3
#define	T3_Def	680.5

#define	Bessel_a	6377397.155
#define	Bessel_1f	299.152813
#define	WGS84_a 	6378137.0
#define	WGS84_1f	298.257223563
#define	PI_S	0.00000484813681109536

#define	maxB	148
#define	maxL	168

int	Par, Rev ;
double	T1, T2, T3 ;
double	dB[maxB][maxL], dL[maxB][maxL], dH[maxB][maxL] ;
double	Bessel_f, Bessel_e2 ;
double	WGS84_f, WGS84_e2 ;

main(argc, argv)
int	argc ;
char	*argv[] ;
{
	FILE	*fp ;
	char	hed, line[500] ;
	long	i, n, nB, nL ;
	double	sB, sL, sH ;
	double	slon, slat, elon, elat, H, H2 ;
	void	tky2wgs(), wgs2tky() ;

	Par = 1 ;
	T1 = T1_Def ;
	T2 = T2_Def ;
	T3 = T3_Def ;
	Rev = 0 ;

	Bessel_f = 1.0 / Bessel_1f ;
	Bessel_e2 = Bessel_f * (2.0 - Bessel_f) ;
	WGS84_f = 1.0 / WGS84_1f ;
	WGS84_e2 = WGS84_f * (2.0 - WGS84_f) ;

	for (i = 1; i < argc; i++) {
	  if (strcmp(argv[i], "-h") == 0) {
	    fprintf(stderr, " USAGE: tky2wgs -pf dsn -3p T1 T2 T3 -r\n") ;
	    return(1) ;
	    }
	  else if (strcmp(argv[i], "-pf") == 0) {
	    Par = 2 ;
	    for (nB = 0; nB < maxB; nB++) {
	      for (nL = 0; nL < maxL; nL++) {
	        dB[nB][nL] = 0.0 ;
	        dL[nB][nL] = 0.0 ;
	        dH[nB][nL] = 0.0 ;
	        }
	      }
	    if ((fp = fopen(argv[++i], "r")) == NULL) {
	      fprintf(stderr, "Parameter file is not found.\n") ;
	      exit(-1) ;
	      }
	    while (1) {
	      if (fgets(line, 100, fp) == NULL) break ;
	      if (line[0] == ' ') line[0] = '0' ;
	      if (line[1] == ' ') line[1] = '0' ;
	      if (line[2] == ' ') line[2] = '0' ;
	      sscanf(line, "%3ld%3ld %lf %lf %lf\n", &nB,&nL,&sB,&sL,&sH) ;
	      if (nB < 0 || nB >= maxB || nL < 0 || nL >= maxL) {
	        fprintf(stderr, "Parameter File Error. %ld %ld\n", nB, nL) ;
	        exit(-1) ;
	        }
	      dB[nB][nL] = sB ;
	      dL[nB][nL] = sL ;
	      dH[nB][nL] = sH ;
	      }
	    fclose(fp) ;
	    }
	  else if (strcmp(argv[i], "-3p") == 0) {
	    Par = 1 ;
	    T1 = atof(argv[++i]) ;
	    T2 = atof(argv[++i]) ;
	    T3 = atof(argv[++i]) ;
	    }
	  else if (strcmp(argv[i], "-r")   == 0) Rev = 1 ;
	  else if (strcmp(argv[i], "-R")   == 0) Rev = 1 ;
	  else {
	    fprintf(stderr, "tky2wgs command argument error.\n") ;
	    exit(-1) ;
	    }
	  }

	H = 0.0 ;

	while ((fgets(line, 500, stdin)) != NULL) {
	  if (line[0] == 'l' || line[0] == 'm') {
	    if (sscanf(line, "%c %lf %lf", &hed, &slon, &slat) != 3) {
	      fprintf(stderr, "JKG l/m data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2wgs(slon, slat, H, &elon, &elat, &H2) ;
	    else wgs2tky(slon, slat, H, &elon, &elat, &H2) ;
	    printf("%c %lf %lf\n", hed, elon, elat) ;
	    }
	  else if(line[0] == 'A' || line[0] == 'W') {
	    if (sscanf(line, "%c %lf %lf", &hed, &slon, &slat) != 3) {
	      fprintf(stderr, "JKG A/W data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2wgs(slon, slat, H, &elon, &elat, &H2) ;
	    else wgs2tky(slon, slat, H, &elon, &elat, &H2) ;
	    printf("%c %lf %lf", hed, elon, elat) ;
	    if (sscanf(line, "%c %*lf %*lf %lf %lf", &hed, &slon, &slat) != 3) {
	      fprintf(stderr, "JKG A/W data error.\n") ;
	      break ;
	      }
	    if (Rev == 0) tky2wgs(slon, slat, H, &elon, &elat, &H2) ;
	    else wgs2tky(slon, slat, H, &elon, &elat, &H2) ;
	    printf(" %lf %lf\n", elon, elat) ;
	    }
	  else printf("%s", line) ;
	  }
	return(0) ;
	}
/* -------------------------------------------------------------------------- */
void	tky2wgs(L, B, H, L2, B2, H2)
double	B, L, H, *B2, *L2, *H2 ;
{
	double	N, X, Y, Z, p, r, mu ;
	int	nB, nL ;
	double	rB, rL, sB, sL, B3, L3, H3, B4, L4, H4, B5, L5, H5 ;

	if (Par == 1) {
	  rB = B * PI_S ;
	  rL = L * PI_S ;
	  N = Bessel_a / sqrt(1.0 - Bessel_e2 * pow(sin(rB),2.0)) ;
	  X = (N + H) * cos(rB) * cos(rL)           + T1 ;
	  Y = (N + H) * cos(rB) * sin(rL)           + T2 ;
	  Z = (N * (1.0 - Bessel_e2) + H) * sin(rB) + T3 ;
	  *L2 = atan(Y / X) / PI_S ;
	  if (X < 0.0) *L2 = *L2 + 648000.0 ;
	  p = sqrt(X*X + Y*Y) ;
	  r = sqrt(p*p + Z*Z) ;
	  mu = atan(Z / p * ((1.0 - WGS84_f) + WGS84_e2*WGS84_a / r)) ;
	  *B2 = atan((Z*(1.0 - WGS84_f) + WGS84_e2*WGS84_a*pow(sin(mu),3.0)) /
		((1.0-WGS84_f) * (p - WGS84_e2*WGS84_a*pow(cos(mu),3.0)))) ;
	  *H2 = p * cos(*B2) + Z * sin(*B2) -
		WGS84_a * sqrt(1.0 - WGS84_e2*pow(sin(*B2),2.0)) ;
	  *B2 = *B2 / PI_S ;
	  }
	else if (Par == 2) {
	  rB = B / 360.0 ;
	  rL = L / 360.0 ;
	  nB = (int)(rB -  310) ;
	  nL = (int)(rL - 1290) ;
	  sB = rB - (double)nB - 310.0 ;
	  sL = rL - (double)nL - 1290.0 ;
	  if (nB < 0 || nB >= maxB-1 || nL < 0 || nL >= maxL-1) {
	    fprintf(stderr, "AREA error. %ld %ld\n", nB, nL) ;
	    exit(-1) ;
	    }
	  if (dH[nB][nL]==0||dH[nB+1][nL]==0||
	    dH[nB+1][nL+1]==0||dH[nB][nL+1]==0) {
	    fprintf(stderr, "AREA error 2. %ld %ld\n", nB, nL) ;
	    exit(-1) ;
	    }
	  B3 = (dB[nB+1][nL]   - dB[nB][nL])   * sB + dB[nB][nL] ;
	  L3 = (dL[nB+1][nL]   - dL[nB][nL])   * sB + dL[nB][nL] ;
	  H3 = (dH[nB+1][nL]   - dH[nB][nL])   * sB + dH[nB][nL] ;
	  B4 = (dB[nB+1][nL+1] - dB[nB][nL+1]) * sB + dB[nB][nL+1] ;
	  L4 = (dL[nB+1][nL+1] - dL[nB][nL+1]) * sB + dL[nB][nL+1] ;
	  H4 = (dH[nB+1][nL+1] - dH[nB][nL+1]) * sB + dH[nB][nL+1] ;
	  B5 = (B4 - B3) * sL + B3 ;
	  L5 = (L4 - L3) * sL + L3 ;
	  H5 = (H4 - H3) * sL + H3 ;
	  *B2 = B + B5 ;
	  *L2 = L + L5 ;
	  *H2 = H + H5 ;
	  }
	}
/* -------------------------------------------------------------------------- */
void	wgs2tky(L, B, H, L2, B2, H2)
double	B, L, H, *B2, *L2, *H2 ;
{
	double	N, X, Y, Z, p, r, mu, rB, rL ;
	double	B3, L3, H3, B4, L4, H4 ;
	void	tky2wgs() ;

	if (Par == 1) {
	  rB = B * PI_S ;
	  rL = L * PI_S ;
	  N = WGS84_a / sqrt(1.0 - WGS84_e2 * pow(sin(rB),2.0)) ;
	  X = (N + H) * cos(rB) * cos(rL)          - T1 ;
	  Y = (N + H) * cos(rB) * sin(rL)          - T2 ;
	  Z = (N * (1.0 - WGS84_e2) + H) * sin(rB) - T3 ;
	  *L2 = atan(Y / X) / PI_S ;
	  if (X < 0.0) *L2 = *L2 + 648000.0 ;
	  p = sqrt(X*X + Y*Y) ;
	  r = sqrt(p*p + Z*Z) ;
	  mu = atan(Z / p * ((1.0 - Bessel_f) + Bessel_e2*Bessel_a / r)) ;
	  *B2 = atan((Z*(1.0-Bessel_f) + Bessel_e2*Bessel_a*pow(sin(mu),3.0)) /
		((1.0-Bessel_f)*(p - Bessel_e2*Bessel_a*pow(cos(mu),3.0)))) ;
	  *H2 = p * cos(*B2) + Z * sin(*B2) -
		Bessel_a * sqrt(1.0 - Bessel_e2*pow(sin(*B2),2.0)) ;
	  *B2 = *B2 / PI_S ;
	  }
	else if (Par == 2) {
	  B3 = B ;
	  L3 = L ;
	  H3 = 0.0 ;
	  tky2wgs(L3, B3, H3, &L4, &B4, &H4) ;
	  B3 = B3 - B4 + B ;
	  L3 = L3 - L4 + L ;
	  tky2wgs(L3, B3, H3, &L4, &B4, &H4) ;
	  *B2 = B3 - B4 + B ;
	  *L2 = L3 - L4 + L ;
	  *H2 = H - H4 ;
	  }
	}
